%%%%%%%%% gradient descend to minimize variance %%%%%%%%%%
clc; clear;
filename = '1M';
load(strcat(filename,'.mat'))
K = 5; % number of ratings
L = 1; % number of latent factors
N_iter = 300; % number of iterations
N_test = N - N_train - N_tune;
rmse_train = zeros(N_iter,1);
rmse_tune = zeros(N_iter,1);
rmse_test = zeros(N_iter,1);
counter = 1;

%%%%%%%%% start iteration %%%%%%%%%%
Lc = ones(L,1)*sqrt(log(hist(lrn_ratings,K))/L);
Ld = Lc;
Lu = zeros(L,K,Nu); % users latent factor matrix
Lm = zeros(L,K,Nm); % movies latent factor matrix
for ii = 1:Nu
    if any(lrn_users == ii)
        Lu(:,:,ii) = (1 - 2*rand(L,K))/10; % users latent factor matrix
    end
end
for ii = 1:Nm
    if any(lrn_movies == ii)
        Lm(:,:,ii) = (1 - 2*rand(L,K))/10; % movies latent factor matrix
    end
end
for ii = 1:N_iter
    ALu = zeros(L,K,N_train);
    ALm = zeros(L,K,N_train);
    for jj = 1:N_train
        ALu(:,:,jj) = Lu(:,:,lrn_users(jj)) + Lc; % augmented users latent factor matrix
        ALm(:,:,jj) = Lm(:,:,lrn_movies(jj)) + Ld; % augmented movies latent factor matrix
    end
    [P,R_bar,rmse] = predict(ALu,ALm,N_train,K,lrn_ratings);
    [dLu,dLm,dLc,dLd] = gradient(P,ALu,ALm,R_bar,lrn_ratings,u2ml,m2ul,N_train,Nu,Nm,K,L); % gradient
    
    s1 = exprnd(1/max(max(max(abs(dLu))))/counter^2/30); % random step size        
    s2 = exprnd(1/max(max(max(abs(dLm))))/counter^2/30); % random step size        
    s3 = exprnd(1/max(max(max(abs(dLc))))/counter^2/30); % random step size        
    s4 = exprnd(1/max(max(max(abs(dLd))))/counter^2/30); % random step size            
    Lu_new = Lu - dLu * s1;
    Lm_new = Lm - dLm * s2;
    Lc_new = Lc - dLc * s3;
    Ld_new = Ld - dLd * s4;
    ALu_new = zeros(L,K,N_train);
    ALm_new = zeros(L,K,N_train);
    for jj = 1:N_train
        ALu_new(:,:,jj) = Lu_new(:,:,lrn_users(jj)) + Lc_new; % augmented users latent factor matrix
        ALm_new(:,:,jj) = Lm_new(:,:,lrn_movies(jj)) + Ld_new; % augmented movies latent factor matrix
    end
    [P_new,R_bar_new,rmse_new] = predict(ALu_new,ALm_new,N_train,K,lrn_ratings);
    if rmse > rmse_new % descend if total square error increases
        Lu = Lu_new;
        Lm = Lm_new;
        Lc = Lc_new;
        Ld = Ld_new;
        counter = 1;
    else
        counter = counter + 1;
    end
    rmse_train(ii) = rmse;

    %%%%%%%%%%%%%%%%% Tuning RMSE %%%%%%%%%%%%%%%%%%%%%%%
    ALu_tune = zeros(L,K,N_tune);
    ALm_tune = zeros(L,K,N_tune);
    for jj = 1:N_tune
        ALu_tune(:,:,jj) = Lu(:,:,tune_users(jj)) + Lc;
        ALm_tune(:,:,jj) = Lm(:,:,tune_movies(jj)) + Ld;
    end    
    [~,~,rmse_tune(ii)] = predict(ALu_tune,ALm_tune,N_tune,K,tune_ratings);

    %%%%%%%%%%%%%%%%% Test RMSE %%%%%%%%%%%%%%%%%%%%%%%
    ALu_test = zeros(L,K,N_test);
    ALm_test = zeros(L,K,N_test);
    for jj = 1:N_test
        ALu_test(:,:,jj) = Lu(:,:,test_users(jj)) + Lc; % augmented users latent factor matrix
        ALm_test(:,:,jj) = Lm(:,:,test_movies(jj)) + Ld; % augmented movies latent factor matrix
    end    
    [~,~,rmse_test(ii)] = predict(ALu_test,ALm_test,N_test,K,test_ratings);
end

save(strcat(filename,'_',num2str(L),'_',num2str(N_iter),'.mat'))